/* Copyright (C) Sean McNamara 2009-2010
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * The Rhythmbox authors hereby grant permission for non-GPL compatible
 * GStreamer plugins to be used and distributed together with GStreamer
 * and Rhythmbox. This permission is above and beyond the permissions granted
 * by the GPL license by which Rhythmbox is covered. If you modify this code
 * you may extend this exception to your version of the code, but you are not
 * obligated to do so. If you do not wish to do so, delete this exception
 * statement from your version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
 */

using GLib;
using RB;
using Gtk;
using Gst;
using GConf;

/*Main class for the plugin, containing the implementation*/
class RBPitchPlugin : RB.Plugin {

    /*The number of seconds to let frequent updates to the adjustments "stack up" before committing them to GConf and the live pipeline.*/
    private static int    STACKUP_TIME = 1;
    
    /*Some math functions for semitone calculations*/
    private static double LN_2      = Math.log(2.0d);
    private static double CONST_S_X = 12.0d / LN_2;
    private static double CONST_G_X = LN_2 / 12.0d;
    
    /*GConf key paths for the settings we persist.*/
    private static string GC_PERSIST_CONTROLS_KEY = "/apps/rhythmbox/plugins/rbpitch/persist-controls";
    private static string GC_PITCH_CTRL_KEY = "/apps/rhythmbox/plugins/rbpitch/pitch-control";
    private static string GC_SPEED_CTRL_KEY = "/apps/rhythmbox/plugins/rbpitch/speed-control";
    private static string GC_TEMPO_CTRL_KEY = "/apps/rhythmbox/plugins/rbpitch/tempo-control";
    private static string GC_PITCH_UB_KEY = "/apps/rhythmbox/plugins/rbpitch/pitch-ubound";
    private static string GC_PITCH_LB_KEY = "/apps/rhythmbox/plugins/rbpitch/pitch-lbound";
    private static string GC_SPEED_UB_KEY = "/apps/rhythmbox/plugins/rbpitch/speed-ubound";
    private static string GC_SPEED_LB_KEY = "/apps/rhythmbox/plugins/rbpitch/speed-lbound";
    private static string GC_TEMPO_UB_KEY = "/apps/rhythmbox/plugins/rbpitch/tempo-ubound";
    private static string GC_TEMPO_LB_KEY = "/apps/rhythmbox/plugins/rbpitch/tempo-lbound";
    private static string GC_RBPITCH_DIR  = "/apps/rhythmbox/plugins/rbpitch";
    
    /*The lower and upper bounds possible for the sliders, in the display format.*/
    private static double PITCH_LB_MIN = xlate_pitch_to_semi(0.1d);
    private static double PITCH_UB_MAX = xlate_pitch_to_semi(10.0d);
    private static double SPEED_LB_MIN = -90.0d;
    private static double SPEED_UB_MAX = 900.0d;
    private static double TEMPO_LB_MIN = -90.0d;
    private static double TEMPO_UB_MAX = 900.0d;
    
    /*The native value to reset to when the reset button is clicked.
      The reset value means the sound is passed through unmodified.*/
    private static double PITCH_RESET  = 0.0d;
    private static double SPEED_RESET  = 0.0d;
    private static double TEMPO_RESET  = 0.0d;
    
    /*Regex parsers for input handler of SpinButtons to convert e.g. "2 st" into double 2.0
      These have to be initialized in activate() because field initializers can't throw errors,
      and the Regex constructor throws a GLib.Error.*/
    private static Regex semitone_input_regex; 
    private static Regex percent_input_regex; 

    /*Used when adjusting the bounds of the sliders.*/
    enum BoundType {
        PITCH_LB, //Lower bound for the pitch control
        PITCH_UB, //Upper bound for the pitch control
        SPEED_LB, //Lower bound for the speed control
        SPEED_UB, //Upper bound for the speed control
        TEMPO_LB, //Lower bound for the tempo control
        TEMPO_UB  //Upper bound for the tempo control
    }


    /*State accounting*/
    private unowned RB.Shell? sh;
    private Gst.Element? pitch_element;
    private static bool inserted = false;
    private uint mergeid = 0;
    private bool persist = true; //Cached value of the persist key in GConf.

    /*The main context source tags for live timeout sources.
      If an id is 0, there is currently no active timeout source for that control.
      If the id is greater than 0, a timeout source is either waiting to expire, or executing right now.
    */
    private uint pitch_adj_timeout_id = 0;
    private uint speed_adj_timeout_id = 0;
    private uint tempo_adj_timeout_id = 0;

    /*For GtkUIManager -- this is a flattened mini XML file defining plugin UI*/
    private static const string ui_addition = 
    """<ui> <toolbar name="ToolBar"> <toolitem name="PitchButton" action="PitchAction"/> </toolbar> </ui>""";

    /*For GtkBuilder -- this is a relative path for the layout file.*/
    private static const string ui_xml = "rbpitch-layout.xml";
    private static const string iconpath = "rbpitchicon.svg"; //Path to button icon

    /*rbpitch Control UI*/
    private Gtk.ToggleAction pitch_action = 
        new Gtk.ToggleAction("PitchAction", null, "Pitch and Tempo Control", null);
    private Gtk.ActionGroup action_group = 
        new Gtk.ActionGroup("PitchActionGroup");
    private Gtk.Image rbpitch_icon = null;
    private unowned Gtk.Window? pitch_ui_parent = null;
    private Gtk.Builder? builder;
    private string? uifile = null;
    private string? iconfile = null;
    private unowned Gtk.Adjustment pitch_adj;
    private unowned Gtk.Adjustment speed_adj;
    private unowned Gtk.Adjustment tempo_adj;
    private unowned Gtk.Button reset_pitch;
    private unowned Gtk.Button reset_tempo;
    private unowned Gtk.Button reset_speed;
    private unowned Gtk.SpinButton pitch_sb;
	private unowned Gtk.SpinButton speed_sb;
    private unowned Gtk.SpinButton tempo_sb;
    
    
    /*RBPlugin-Config UI*/
    private unowned Gtk.Window? config_ui_parent = null;
    private unowned Gtk.Adjustment adj_config_tempo_lb;
    private unowned Gtk.Adjustment adj_config_tempo_ub;
    private unowned Gtk.Adjustment adj_config_speed_lb;
    private unowned Gtk.Adjustment adj_config_speed_ub;
    private unowned Gtk.Adjustment adj_config_pitch_lb;
    private unowned Gtk.Adjustment adj_config_pitch_ub;
    private unowned Gtk.SpinButton sb_config_pitch_lb;
    private unowned Gtk.SpinButton sb_config_pitch_ub;
    private unowned Gtk.SpinButton sb_config_speed_lb;
    private unowned Gtk.SpinButton sb_config_speed_ub;
    private unowned Gtk.SpinButton sb_config_tempo_lb;
    private unowned Gtk.SpinButton sb_config_tempo_ub;
    private unowned Gtk.Button     button_config_close;
    private unowned Gtk.CheckButton check_config_persist;
            
    /*GConf data*/
    private unowned GConf.Client? gconf = null;
    
    /*NEVER access these directly outside of the property methods!!*/
    private double _drate  = 1.0f;
    private double _dtempo = 1.0f;
    private double _dpitch = 1.0f;

    /*Corresponds to the 'rate' property of the Gstreamer plugin.
      Valid values are 0.1 - 10.0*/
    public double drate {
        get {
            return _drate;
        }

        set {
            if(value < 0.1d || value > 10.0d) {
                GLib.critical("drate value out of bounds: %f\n", value);
                return;
            }
            _drate = value;
            if(pitch_element != null) {
                pitch_element.set("rate", _drate);
            }
        }
    }

    /*Corresponds to the 'pitch' property of the Gstreamer plugin.
      Valid values are 0.1 - 10.0*/
    public double dpitch {
        get {
            return _dpitch;
        }

        set {
            if(value < 0.1d || value > 10.0d) {
                GLib.critical("dpitch value out of bounds: %f\n", value);
                return;
            }
            _dpitch = value;
            if(pitch_element != null) {
                pitch_element.set("pitch", _dpitch);
            }
        }
    }

    /*Corresponds to the 'tempo' property of the Gstreamer plugin.
      Valid values are 0.1 - 10.0*/
    public double dtempo {
        get {
            return _dtempo;
        }

        set {
            if(value < 0.1d || value > 10.0d) {
                GLib.critical("dtempo value out of bounds: %f\n", value);
                return;
            }

            _dtempo = value;
            if(pitch_element != null) {
                pitch_element.set("tempo", _dtempo);
            }
        }
    }
    
    public RBPitchPlugin() {
        GLib.Object();
    }


    /*The plugin activation hook. This is the first function called by Rhythmbox
      when it opens our dynamically linked plugin.*/
    public override void activate (RB.Shell shell) {
	    this.sh = shell;
	    
	    RBPitchPlugin.semitone_input_regex = new Regex("""(?P<whole>-?((\d+\.\d+)|(\.\d+)|(\d+)))\s*(st|s\.t\.)?""", RegexCompileFlags.CASELESS | RegexCompileFlags.OPTIMIZE, RegexMatchFlags.NOTEMPTY);
	    
	    RBPitchPlugin.percent_input_regex = new Regex("""(?P<whole>-?((\d+\.\d+)|(\.\d+)|(\d+)))\s*%?""", RegexCompileFlags.CASELESS | RegexCompileFlags.OPTIMIZE, RegexMatchFlags.NOTEMPTY);

	    this.uifile = this.find_file(ui_xml);
	    this.iconfile = this.find_file(iconpath);
	    if(this.uifile == null) {
	        no_ui_file_error();
	        return;
	    }
	
	    if(!have_gst_player()) {
	        no_gst_error();
	        return;
	    }

	    if(shell == null || !pitch_init_check(true)) {
	        pitch_not_found_error();
	        return;
	    }
	
	    insert_pitch_element();

	    //Rhythmbox UI signal for the toggle button to show/hide the toplevel
	    pitch_action.activate.connect((t) => {
	        if(((Gtk.ToggleAction) t).get_active()) {
	            //The button is toggled on, so show the UI.
	            display_ui();
	        }
	        else {
	            //The button is toggled off, so hide (not destroy!) the UI.
	            hide_ui();
	        }
	    });   
      
	    //These few lines add our plugin button to the Rhythmbox shell.
	    rbpitch_icon = new Gtk.Image.from_file(iconfile);
	    action_group.add_action(pitch_action);
	    Gtk.UIManager uim = ((Gtk.UIManager) shell.get_ui_manager());
	    uim.insert_action_group(action_group, 0);
	    this.mergeid = uim.add_ui_from_string(ui_addition, -1);
	    uim.ensure_update();

	    this.builder = new Gtk.Builder();
	    this.builder.add_from_file(uifile);
		    if(!gconf_init()) {
			    no_gconf_error();
		    }
	    construct_config_ui();
	    construct_ui();
	    
	    //Set the custom icon
   	    Gtk.Widget? mwidge = uim.get_widget("/ui/ToolBar/PitchButton");
   	    if(mwidge == null)
   	    {
   	        GLib.critical("Invalid widget path in uim.get_widget()\n");
   	    }
   	    else
   	    {
   	        //GLib.warning("Type name of mwidge: " + mwidge.get_type().name() + "\n");
   	        ((Gtk.ToggleToolButton) mwidge).set_icon_widget(rbpitch_icon);
   	        ((Gtk.ToggleToolButton) mwidge).set_label_widget(rbpitch_icon);
   	        rbpitch_icon.show();
   	    }
    }

    /*The plugin deactivation hook. Called when the plugin is asked to unload.*/
	public override void deactivate (RB.Shell shell) {
		this.sh = shell;
		remove_pitch_element();
		destroy_ui();
		destroy_config_ui();
		
		//This removes our plugin button from the Rhythmbox shell.
		Gtk.UIManager uim = ((Gtk.UIManager) shell.get_ui_manager());
		uim.remove_ui(this.mergeid);
		uim.remove_action_group(this.action_group);
		action_group.remove_action(pitch_action);
		this.pitch_action.set_active(false);
		this.sh = null;
		this.uifile = null;
	}
	
	/*Plugin configuration UI, using Rhythmbox's shell plugin config hook*/
	private void construct_config_ui() {
		if(builder == null) {
			this.builder = new Gtk.Builder();
			this.builder.add_from_file(uifile);
		}
		
		config_ui_parent = (Gtk.Window) this.builder.get_object("window-config");
		config_ui_parent.delete_event.connect((t, a) => {
		    hide_config_ui();
		    return true;
		});

        adj_config_pitch_ub  = (Gtk.Adjustment)  this.builder.get_object("adjustment-config-pitch-ub");		
        adj_config_pitch_lb  = (Gtk.Adjustment)  this.builder.get_object("adjustment-config-pitch-lb");
        adj_config_speed_lb  = (Gtk.Adjustment)  this.builder.get_object("adjustment-config-speed-lb");
        adj_config_speed_ub  = (Gtk.Adjustment)  this.builder.get_object("adjustment-config-speed-ub");
        adj_config_tempo_lb  = (Gtk.Adjustment)  this.builder.get_object("adjustment-config-tempo-lb");
        adj_config_tempo_ub  = (Gtk.Adjustment)  this.builder.get_object("adjustment-config-tempo-ub");
        sb_config_pitch_ub   = (Gtk.SpinButton)  this.builder.get_object("sb-config-pitch-ub");
        sb_config_pitch_lb   = (Gtk.SpinButton)  this.builder.get_object("sb-config-pitch-lb");
        sb_config_speed_ub   = (Gtk.SpinButton)  this.builder.get_object("sb-config-speed-ub");
        sb_config_speed_lb   = (Gtk.SpinButton)  this.builder.get_object("sb-config-speed-lb");
        sb_config_tempo_ub   = (Gtk.SpinButton)  this.builder.get_object("sb-config-tempo-ub");
        sb_config_tempo_lb   = (Gtk.SpinButton)  this.builder.get_object("sb-config-tempo-lb");
        button_config_close  = (Gtk.Button)      this.builder.get_object("button-config-close");
        check_config_persist = (Gtk.CheckButton) this.builder.get_object("check-config-persist");

        adj_config_pitch_ub.value_changed.connect((t) => {
			gconf.set_float(GC_PITCH_UB_KEY, t.get_value());        	
        });
        
        adj_config_pitch_lb.value_changed.connect((t) => {
			gconf.set_float(GC_PITCH_LB_KEY, t.get_value());        	
        });
        
        adj_config_speed_ub.value_changed.connect((t) => {
			gconf.set_float(GC_SPEED_UB_KEY, t.get_value());        	
        });
        
        adj_config_speed_lb.value_changed.connect((t) => {
			gconf.set_float(GC_SPEED_LB_KEY, t.get_value());        	
        });
        
        adj_config_tempo_ub.value_changed.connect((t) => {
			gconf.set_float(GC_TEMPO_UB_KEY, t.get_value());        	
        });
        
        adj_config_tempo_lb.value_changed.connect((t) => {
			gconf.set_float(GC_TEMPO_LB_KEY, t.get_value());        	
        });
        
        button_config_close.clicked.connect((t) => {
        	hide_config_ui();
        });
        
        check_config_persist.toggled.connect((t) => {
	        gconf.set_bool(GC_PERSIST_CONTROLS_KEY, check_config_persist.get_active());
        });

        /*Output signals format the GtkSpinButton values with "%" or "st"*/        
        sb_config_pitch_ub.output.connect(RBPitchPlugin.semitone_output_cb);
        sb_config_pitch_lb.output.connect(RBPitchPlugin.semitone_output_cb);
        sb_config_speed_ub.output.connect(RBPitchPlugin.percent_output_cb);
        sb_config_speed_lb.output.connect(RBPitchPlugin.percent_output_cb);
        sb_config_tempo_ub.output.connect(RBPitchPlugin.percent_output_cb);
        sb_config_tempo_lb.output.connect(RBPitchPlugin.percent_output_cb);

        /*Input signals attempt to convert typed GtkSpinButton values into numbers by stripping off % or st*/
        
        sb_config_pitch_ub.input.connect(RBPitchPlugin.semitone_input_cb);
        sb_config_pitch_lb.input.connect(RBPitchPlugin.semitone_input_cb);
        sb_config_speed_ub.input.connect(RBPitchPlugin.percent_input_cb);
        sb_config_speed_lb.input.connect(RBPitchPlugin.percent_input_cb);
        sb_config_tempo_ub.input.connect(RBPitchPlugin.percent_input_cb);
        sb_config_tempo_lb.input.connect(RBPitchPlugin.percent_input_cb);
        
        download_bounds();
	}
	
	private void download_bounds() {
		adj_config_pitch_ub.value = gconf.get_float(GC_PITCH_UB_KEY);
		adj_config_pitch_lb.value = gconf.get_float(GC_PITCH_LB_KEY);
		adj_config_speed_ub.value = gconf.get_float(GC_SPEED_UB_KEY);
		adj_config_speed_lb.value = gconf.get_float(GC_SPEED_LB_KEY);
		adj_config_tempo_ub.value = gconf.get_float(GC_TEMPO_UB_KEY);
		adj_config_tempo_lb.value = gconf.get_float(GC_TEMPO_LB_KEY);
	}

    /*Build our plugin UI in its own toplevel; this has nothing to do with rb-ui*/
    private void construct_ui() {
        pitch_ui_parent = (Gtk.Window) this.builder.get_object("window-control");
        pitch_ui_parent.delete_event.connect((t, a) => {
            hide_ui();
            this.pitch_action.set_active(false);
            return true;
        });

        //Pull the controls that need code into private class variables.
        pitch_adj   = (Gtk.Adjustment) this.builder.get_object("adjustment-control-pitch");
        speed_adj   = (Gtk.Adjustment) this.builder.get_object("adjustment-control-speed");
        tempo_adj   = (Gtk.Adjustment) this.builder.get_object("adjustment-control-tempo");
        reset_pitch = (Gtk.Button)     this.builder.get_object("button-reset-pitch");
        reset_tempo = (Gtk.Button)     this.builder.get_object("button-reset-tempo");
        reset_speed = (Gtk.Button)     this.builder.get_object("button-reset-speed");
	    pitch_sb    = (Gtk.SpinButton) this.builder.get_object("spinbutton-control-pitch");
    	speed_sb    = (Gtk.SpinButton) this.builder.get_object("spinbutton-control-speed");
	    tempo_sb    = (Gtk.SpinButton) this.builder.get_object("spinbutton-control-tempo");

        /*
        Implementation Note:
        Our GtkAdjustment has a range of 10 ~ 1000.
        The pitch element in Gstreamer has a range of 0.1 ~ 10.0.
        We have to scale between the user value (percents) and the internal value.
        Using the user value in the adjustment lets us use the same adjustment
        for both the SpinButton and the HScale, keeping them in sync automatically
        and displaying a sane value to the user. Then we just convert to the
        internal representation in the signal handlers below.
        Note also that the adjustments have an update policy of DELAYED,
        so these signal handlers won't be called frantically when you drag the
        HScale. So the double precision division won't be a big performance hit.
        */

        pitch_adj.value_changed.connect((t) => {
            if(pitch_adj_timeout_id == 0) {
                pitch_adj_timeout_id = GLib.Timeout.add_seconds(STACKUP_TIME, this.pitch_adj_timeout_cb);
            }
        });
        
        speed_adj.value_changed.connect((t) => {
            if(speed_adj_timeout_id == 0) {
                GLib.Timeout.add_seconds(STACKUP_TIME, this.speed_adj_timeout_cb);
            }
        });

        tempo_adj.value_changed.connect((t) => {
            if(tempo_adj_timeout_id == 0) {
                GLib.Timeout.add_seconds(STACKUP_TIME, this.tempo_adj_timeout_cb);
            }
        });

        reset_pitch.clicked.connect((t) => {
        	double resetVal = PITCH_RESET;
        	clamp(ref resetVal, get_pitch_lb(), get_pitch_ub());
            pitch_adj.set_value(resetVal);
            if(persist) {
	            gconf.set_float(GC_PITCH_CTRL_KEY, resetVal);
            }
        });

        reset_speed.clicked.connect((t) => {
        	double resetVal = SPEED_RESET;
        	clamp(ref resetVal, get_speed_lb(), get_speed_ub());
            speed_adj.set_value(resetVal);
            if(persist) {
	            gconf.set_float(GC_SPEED_CTRL_KEY, resetVal);
            }
        });

        reset_tempo.clicked.connect((t) => {
        	double resetVal = TEMPO_RESET;
        	clamp(ref resetVal, get_tempo_lb(), get_tempo_ub());
            tempo_adj.set_value(resetVal);
            if(persist) {
	            gconf.set_float(GC_TEMPO_CTRL_KEY, resetVal);
            }
        });

	    pitch_sb.output.connect(RBPitchPlugin.semitone_output_cb);
	    speed_sb.output.connect(RBPitchPlugin.percent_output_cb);
	    tempo_sb.output.connect(RBPitchPlugin.percent_output_cb);
	    
	    pitch_sb.input.connect(RBPitchPlugin.semitone_input_cb);
	    speed_sb.input.connect(RBPitchPlugin.percent_input_cb);
	    tempo_sb.input.connect(RBPitchPlugin.percent_input_cb);
	    
	    download_controls();
    }
    
    public bool pitch_adj_timeout_cb() {
        double val = pitch_adj.get_value();
        dpitch = xlate_pitch_to_native(val);
        if(persist) {
	        gconf.set_float(GC_PITCH_CTRL_KEY, val);
        }
        pitch_adj_timeout_id = 0;
        return false;
    }
    
    public bool speed_adj_timeout_cb() {
        double val = speed_adj.get_value();
        drate = xlate_ts_to_native(val);
        if(persist) {
            gconf.set_float(GC_SPEED_CTRL_KEY, val);
        }
        speed_adj_timeout_id = 0;
        return false;
    }
    
    public bool tempo_adj_timeout_cb() {
    	double val = tempo_adj.get_value();
        dtempo = xlate_ts_to_native(val);
        if(persist) {
            gconf.set_float(GC_TEMPO_CTRL_KEY, val);
        }
        tempo_adj_timeout_id = 0;
        return false;
    }
    
    private static int generic_input_cb(Gtk.SpinButton t, void* new_value, Regex reg) {
        bool success = false;
        double* newval = (double*) new_value; //It's really a double*
        string btn_txt = t.get_text();
        double from_spinbutton = 0.0;
        MatchInfo mi;
        
        success = reg.match(btn_txt, 0, out mi);
        string? number = mi.fetch_named("whole");
        if(number == null) {
            success = false;
        }
        else {
            from_spinbutton = number.to_double();
        }
        
        if(success) {
            *newval = from_spinbutton;
            return (int) true;
        }
        else {
            return Gtk.INPUT_ERROR;
        }
    }
    
    public static int semitone_input_cb(Gtk.SpinButton t, void* new_value) {
        return generic_input_cb(t, new_value, semitone_input_regex);
    }
    
    public static int percent_input_cb(Gtk.SpinButton t, void* new_value) {
        return generic_input_cb(t, new_value, percent_input_regex);
    }
    
    public static bool semitone_output_cb(Gtk.SpinButton t) {
    	double doub = t.get_adjustment().get_value();
    	string s = "%.2f st";
    	s.printf(doub);
        t.set_text(s.printf(doub));
        return true;
    }
    
    public static bool percent_output_cb(Gtk.SpinButton t) {
    	double doub = t.get_adjustment().get_value();
    	string s = "%.1f%%";
        t.set_text(s.printf(doub));
        return true;
    }
    
    private void download_controls() {
    	if(persist) {
			pitch_adj.value = gconf.get_float(GC_PITCH_CTRL_KEY);
			speed_adj.value = gconf.get_float(GC_SPEED_CTRL_KEY);
			tempo_adj.value = gconf.get_float(GC_TEMPO_CTRL_KEY);
		}
    }

    /*Destroy the UI permanently (until activate is called again).
    Does NOT handle our Rhythmbox shell additions.*/
    private void destroy_ui() {
    	if(pitch_ui_parent != null) {
	        pitch_ui_parent.destroy();
    	    pitch_ui_parent = null;
    	}
    }
    
    private void destroy_config_ui() {
    	if(config_ui_parent != null) {
	    	config_ui_parent.destroy();
	    	config_ui_parent = null;
	   	}
    }

    /*Display the plugin control toplevel (not related to the Rhythmbox Shell).*/
    private void display_ui() {
        if(pitch_ui_parent != null)
            pitch_ui_parent.present();
    }

    /*Hide the plugin control toplevel.*/
    private void hide_ui() {
        pitch_ui_parent.hide();
    }
    
    /*Hide the plugin config toplevel.*/
    private void hide_config_ui() {
        config_ui_parent.hide();
    }

    /*
    Inserts an element of the class "pitch" in the Gstreamer pipeline as a filter.
    */
    private void insert_pitch_element() {
        if(!inserted && pitch_init_check(true) && sh != null)
        {
            RB.ShellPlayer? obj = sh.get_player() as RB.ShellPlayer;
            if(obj != null) {
                unowned RB.PlayerGstFilter? filt = get_filter();
                if(filt != null) {
                    filt.add_filter(pitch_element);
                    inserted = true;
                }
            }
            else {
                GLib.critical("ShellPlayer was NULL\n");
            }
        }
        else {
            GLib.critical("inserted || !pitch_init_check() || sh == NULL\n");
        }
    }

    /*Reverses the operations of insert_pitch_element, also stopping playback.*/
    private void remove_pitch_element() {
        if(inserted && pitch_element != null && sh != null)
        {
            RB.ShellPlayer? obj = sh.get_player() as RB.ShellPlayer;
            if(obj != null) {
                unowned RB.PlayerGstFilter? filt = get_filter();
                if(filt != null) {
                    filt.remove_filter(pitch_element);
                    inserted = false;
                }
            }
            else {
                GLib.critical("ShellPlayer was NULL\n");
            }
        }
        else {
            GLib.critical("!inserted || pitch_element == NULL || sh == NULL\n");
        }
    }

    /*Returns the shell's RBPlayer instance casted to the PlayerGstFilter interface.*/
    private unowned RB.PlayerGstFilter? get_filter() {
        if(sh == null)
            return null;

        unowned RB.ShellPlayer? obj = sh.get_player() as RB.ShellPlayer;
        if(obj == null) {
            return null;
        }

        unowned GLib.Object? v = null;
        obj.get("player", &v);
        if(v is RB.PlayerGstFilter) {
            return (RB.PlayerGstFilter) v;
        }
        else {
            GLib.critical("RBPlayer does not implement PlayerGstFilter interface\n");
            return null;
        }
    }

    /*Returns the shell's RBPlayer instance.*/
    private unowned RB.Player? get_player_backend() {
        RB.ShellPlayer? obj = sh.get_player() as RB.ShellPlayer;
        if(obj == null) {
            return null;
        }

        unowned GLib.Object? v = null;
        obj.get("player", &v);
        if(v is RB.Player) {
            return (RB.Player) v;
        }
        else {
            GLib.critical("RBShellPlayer's player property is not an RBPlayer!\n");
            return null;
        }
    }

    /*Performs RTTI on the RBPlayer to determine if it uses GStreamer.*/
    public bool have_gst_player() {
        return get_player_backend() is RB.PlayerGstFilter;
    }

    /*Tries to instantiate a pitch element and returns false if it can't.*/
    public bool pitch_init_check(bool force = false) {
        if(force || pitch_element != null) {
            this.pitch_element = Gst.ElementFactory.make("pitch", null);
            if(pitch_element != null) {
                pitch_element.set("rate", this.drate);
                pitch_element.set("pitch", this.dpitch);
                pitch_element.set("tempo", this.dtempo);                
            }
        }
        return (this.pitch_element != null);
    }

    public bool gconf_init() {
        gconf = GConf.Client.get_default();
        
        if(gconf == null) {
        	return false;
        }
        
        //If the schema didn't get installed, set widest possible range as default
        if(!gconf.dir_exists(GC_RBPITCH_DIR)) {
        	gconf.set_float(GC_PITCH_UB_KEY, PITCH_UB_MAX);
        	gconf.set_float(GC_PITCH_LB_KEY, PITCH_LB_MIN);
        	gconf.set_float(GC_SPEED_UB_KEY, SPEED_UB_MAX);
        	gconf.set_float(GC_SPEED_LB_KEY, SPEED_LB_MIN);
        	gconf.set_float(GC_TEMPO_UB_KEY, TEMPO_UB_MAX);
        	gconf.set_float(GC_TEMPO_LB_KEY, TEMPO_LB_MIN);
        	gconf.set_float(GC_PITCH_CTRL_KEY, PITCH_RESET);
        	gconf.set_float(GC_SPEED_CTRL_KEY, SPEED_RESET);
        	gconf.set_float(GC_TEMPO_CTRL_KEY, TEMPO_RESET);
        	gconf.set_bool(GC_PERSIST_CONTROLS_KEY, true);
        }
        
        gconf.add_dir(GC_RBPITCH_DIR, GConf.ClientPreloadType.ONELEVEL);
        gconf.notify_add(GC_RBPITCH_DIR, (t, cnxn_id, entry) => {
            unowned string k = entry.get_key();
            unowned GConf.Value v = entry.get_value();
            unowned GConf.ValueType vt = v.type;
            
            if(vt == GConf.ValueType.FLOAT) {
		        double f = v.get_float();

		        //Pitch upper bound
		        if(k == GC_PITCH_UB_KEY) {
		            try_adjust_bound(f, BoundType.PITCH_UB);
		            return;
		        }

		        //Pitch lower bound
		        if(k == GC_PITCH_LB_KEY) {
		            try_adjust_bound(f, BoundType.PITCH_LB);
		            return;
		        }

		        //Speed upper bound
		        if(k == GC_SPEED_UB_KEY) {
		            try_adjust_bound(f, BoundType.SPEED_UB);
		            return;
		        }

		        //Speed lower bound
		        if(k == GC_SPEED_LB_KEY) {
		            try_adjust_bound(f, BoundType.SPEED_LB);
		            return;
		        }

		        //Tempo upper bound
		        if(k == GC_TEMPO_UB_KEY) {
		            try_adjust_bound(f, BoundType.TEMPO_UB);
		            return;
		        }

		        //Tempo lower bound
		        if(k == GC_TEMPO_LB_KEY) {
		            try_adjust_bound(f, BoundType.TEMPO_LB);
		            return;
		        }
		        
		        if(k == GC_PITCH_CTRL_KEY && persist) {
		        	pitch_adj.set_value(f);
	        	}
	        	
	        	if(k == GC_SPEED_CTRL_KEY && persist) {
	        		speed_adj.set_value(f);
	        	}
	        	
	        	if(k == GC_TEMPO_CTRL_KEY && persist) {
	        		tempo_adj.set_value(f);
	        	}
	        }
	        
	        if(vt == GConf.ValueType.BOOL) {
	        	bool b = v.get_bool();
		        if(k == GC_PERSIST_CONTROLS_KEY) {
		        	persist = b;
		    	}
        	}
        });
        
        persist = gconf.get_bool(GC_PERSIST_CONTROLS_KEY);
        
        return true;
    }
    
    private double get_pitch_lb() {
    	return gconf.get_float(GC_PITCH_LB_KEY);
    }
    
    private double get_pitch_ub() {
    	return gconf.get_float(GC_PITCH_UB_KEY);
    }
    
    private double get_speed_lb() {
    	return gconf.get_float(GC_SPEED_LB_KEY);
    }
    
    private double get_speed_ub() {
    	return gconf.get_float(GC_SPEED_UB_KEY);
    }
    
    private double get_tempo_lb() {
    	return gconf.get_float(GC_TEMPO_LB_KEY);
    }
    
    private double get_tempo_ub() {
    	return gconf.get_float(GC_TEMPO_UB_KEY);
    }

    private bool try_adjust_bound(double val, BoundType bt) {
        switch(bt) {
            case BoundType.PITCH_UB:
				return try_adjust_bound_priv(val, gconf.get_float(GC_PITCH_LB_KEY), PITCH_UB_MAX, GC_PITCH_UB_KEY, GC_PITCH_CTRL_KEY, adj_config_pitch_ub, pitch_adj, true);
			case BoundType.PITCH_LB:
				return try_adjust_bound_priv(val, PITCH_LB_MIN, gconf.get_float(GC_PITCH_UB_KEY), GC_PITCH_LB_KEY, GC_PITCH_CTRL_KEY, adj_config_pitch_lb, pitch_adj, false);
            case BoundType.SPEED_UB:
				return try_adjust_bound_priv(val, gconf.get_float(GC_SPEED_LB_KEY), SPEED_UB_MAX, GC_SPEED_UB_KEY, GC_SPEED_CTRL_KEY, adj_config_speed_ub, speed_adj, true);
			case BoundType.SPEED_LB:
				return try_adjust_bound_priv(val, SPEED_LB_MIN, gconf.get_float(GC_SPEED_UB_KEY), GC_SPEED_LB_KEY, GC_SPEED_CTRL_KEY, adj_config_speed_lb, speed_adj, false);
            case BoundType.TEMPO_UB:
				return try_adjust_bound_priv(val, gconf.get_float(GC_TEMPO_LB_KEY), TEMPO_UB_MAX, GC_TEMPO_UB_KEY, GC_TEMPO_CTRL_KEY, adj_config_tempo_ub, tempo_adj, true);
			case BoundType.TEMPO_LB:
				return try_adjust_bound_priv(val, TEMPO_LB_MIN, gconf.get_float(GC_TEMPO_UB_KEY), GC_TEMPO_LB_KEY, GC_TEMPO_CTRL_KEY, adj_config_tempo_lb, tempo_adj, false);
			default:
				return false;
        }
    }
    
    private bool try_adjust_bound_priv(double newval, double oldlow, double oldhigh, string config_key, string ctrl_key, Gtk.Adjustment config_adj, Gtk.Adjustment control_adj, bool isUpper) {
    	if(clamp(ref newval, oldlow, oldhigh)) {
    		gconf.set_float(config_key, newval);
    		return false;
    	}
    	else {
    		config_adj.value = newval;
    		double ctrl_val = control_adj.get_value();
    		if(isUpper) { //newval is the new upper bound and oldlow is still the lower bound
	    		clamp(ref ctrl_val, oldlow, newval);
	    		control_adj.set_upper(newval);
	    	}
	    	else { //newval is the new lower bound and oldhigh is still the upper bound
	    		clamp(ref ctrl_val, newval, oldhigh);
	    		control_adj.set_lower(newval);
	    	}
	    	control_adj.set_value(ctrl_val);
	    	if(persist) {
		    	gconf.set_float(ctrl_key, ctrl_val);
	    	}
    		return true;
    	}
    }
    
    public override unowned Gtk.Widget create_configure_dialog () {
    	return config_ui_parent;
    }

    /*Displays a message dialog explaining what the user has to do to use the
      plugin in case they don't have libgstsoundtouch.*/
    private void pitch_not_found_error() {
        display_error("The pitch plugin was not found.\nPlease install GStreamer" +
        "-plugins-bad-0.10 with SoundTouch support.\nUnload the Pitch and Tempo Shifting plugin to stop" +
        " this message from appearing.");
    }

    /*Displays a message dialog explaining that we can't proceed because the
      player backend is not Gstreamer.*/
    private void no_gst_error() {
        display_error("Pitch and Tempo Shifting can not be used with your configuration." +
        "\nPlease use the GStreamer-0.10 backend.\nUnload the Pitch and Tempo Shifting plugin to stop" +
        " this message from appearing.");
    }
    
    private void no_gconf_error() {
    	display_error("Pitch and Tempo Shifting can not be used without GConf.\nUnload the Pitch and Tempo Shifting plugin to stop" +
        " this message from appearing.");
    }

    /*Displays a message dialog explaining that we can't proceed because we can't
      find the GtkBuilder UI file (usually permissions or directory path PEBKAC)*/
    private void no_ui_file_error() {
        display_error("The Pitch plugin can not find its UI file.\nPlease check " +
        "your configuration.\nUnload the Pitch and Tempo Shifting plugin to stop" +
        " this message from appearing.");
    }

    /*Convenience method to show an error dialog.*/
    private void display_error(string msg) {
        Gtk.MessageDialog dialog = new Gtk.MessageDialog(null,
                                  DialogFlags.DESTROY_WITH_PARENT,
                                  Gtk.MessageType.ERROR,
                                  ButtonsType.CLOSE,
                                  msg);
        dialog.run();
        dialog.destroy();
    }
    
    /*Translate a pitch value from libgstsoundtouch native values
      to "Equal 12-tone temperament semitones".
      The domain of this function is 0.1 ~ 10.0, and
      the range is approx. -39.8 ~ 39.8.
      This function is the mathematical inverse of xlate_pitch_to_native().
      Example: xlate_pitch_to_semi(2.0) returns 12.0, which is +12 Semitones.
      */
    private static inline double xlate_pitch_to_semi(double pitch_native) {
        clamp(ref pitch_native, 0.1d, 10.0d);
        return CONST_S_X * Math.log(pitch_native);
    }
    
    /*Translate a pitch value from "Equal 12-tone temperament semitones"
      to libgstsoundtouch native values.
      The domain of this function is approx. -39.8 ~ 39.8, and
      the range is 0.1 ~ 10.0.
      This function is the mathematical inverse of xlate_pitch_to_semi().
      Example: xlate_pitch_to_native(12.0) returns 2.0.
      */
    private static inline double xlate_pitch_to_native(double pitch_semi) {
    	clamp(ref pitch_semi, xlate_pitch_to_semi(0.1d), xlate_pitch_to_semi(10.0d));
        return Math.exp(pitch_semi * CONST_G_X);
    }
    
    /*Translate a tempo or speed value from libgstsoundtouch native values
      to percentage-deltas.
      The domain of this function is 0.1 ~ 10.0, and
      the range is -90% ~ +900%.
      This function is the mathematical inverse of xlate_ts_to_native().
      Example: xlate_ts_to_semi(2.0) returns 100.0, which is +100% increase.
      */
    private static inline double xlate_ts_to_semi(double ts_native) {
        clamp(ref ts_native, 0.1d, 10.0d);
        return 100.0d * (ts_native - 1.0d);
    }
    
    
    /*Translate a tempo or speed value from percentage-deltas
    to libgstsoundtouch native values.
    The domain of this function is -90.0 ~ 900.0, and
    the range is 0.1 ~ 10.0.
    This function is the mathematical inverse of xlate_ts_to_semi().
    Example: xlate_ts_to_native(100.0) returns 2.0.
    */
    private static inline double xlate_ts_to_native(double ts_semi) {
    	clamp(ref ts_semi, -90.0d, 900.0d);
        return 1.0d + (ts_semi / 100.0d);
    }
    
    /*Returns true if the value was clamped, false otherwise.*/
    private static bool clamp(ref double test, double low, double high) {
    	if(test < low) {
    		test = low;
    		return true;
    	}
    	else {
    		if(test > high) {
    			test = high;
    			return true;
    		}
    	}
    	return false;
    }
}

/*Required GLib boilerplate code.*/
[ModuleInit]
public GLib.Type register_rb_plugin (GLib.TypeModule module)
{
	return typeof (RBPitchPlugin);
}
